"""Email delivery checkers for messenger."""
import logging
from datetime import datetime
from email import message_from_string
from typing import Any, Dict, Optional, Union, cast

from imap_tools import AND, MailBox, MailBoxUnencrypted, MailMessage
from pytz import UTC
from sqlalchemy.orm import Session
from sqlalchemy.orm.exc import NoResultFound

from ..constants import ArchiveMessageStatus, MessageStatus
from ..database import atomic
from ..models import ArchiveEmail, Email
from ..utils import create_ssl_context

try:
    from imap_tools import MailMessageFlags
except ImportError:  # pragma: no cover
    # Fallback for imap_tools < 0.18
    from imap_tools import MessageFlags as MailMessageFlags

_LOGGER = logging.getLogger(__name__)
MIMETYPE_REPORT = 'multipart/report'
MIMETYPE_MESSAGE = 'message/rfc822'
MIMETYPE_MESSAGE_HEADERS = 'text/rfc822-headers'
MIMETYPE_DSN = 'message/delivery-status'
REPORT_TYPE_DSN = 'delivery-status'
DSN_FIELD_NAMES = (
    # Message fields
    'Reporting-MTA',
    # Recipient fields
    'Action', 'Status', 'Remote-MTA', 'Diagnostic-Code', 'Final-Recipient',
)


class ImapDelivery:
    """Check email delivery from mailbox using IMAP."""

    def __init__(self, host: str = '', port: int = None, ssl: bool = True, tls: bool = False,
                 keyfile: str = None, certfile: str = None, username: str = None, password: str = None):
        self.host = host
        self.port = port
        self.ssl = ssl
        self.tls = tls
        self.keyfile = keyfile
        self.certfile = certfile
        self.username = username
        self.password = password

    def get_mailbox(self) -> MailBox:
        """Get a mailbox instance."""
        if self.ssl:
            mailbox = MailBox(host=self.host, port=self.port,
                              ssl_context=create_ssl_context(self.certfile, self.keyfile))
        else:
            mailbox = MailBoxUnencrypted(host=self.host, port=self.port)

        if self.tls:
            mailbox.box.starttls(create_ssl_context(self.certfile, self.keyfile))
        if self.username and self.password:
            mailbox.login(username=self.username, password=self.password)
        return mailbox

    def is_dsn(self, message: MailMessage) -> bool:
        """Return True if message is a delivery status."""
        return bool(message.obj.get_content_type() == MIMETYPE_REPORT
                    and message.obj.get_param('report-type') == REPORT_TYPE_DSN)

    def get_status(self, message: MailMessage) -> Optional[MessageStatus]:
        """Return an email status based on a message content."""
        if message.obj.get_content_type() != MIMETYPE_REPORT:
            # It's a reply, it's been delivered.
            return MessageStatus.DELIVERED
        if not self.is_dsn(message):
            # Not a valid delivery status
            return None
        # Collect all Action fields
        actions = [fields['Action'] for part in message.obj.get_payload()
                   for fields in part.get_payload() if 'Action' in fields]
        if not actions:
            return None
        # XXX: Just use the first one, and ignore the rest for now.
        action = actions[0]
        if action == 'failed':
            return MessageStatus.UNDELIVERED
        elif action == 'delivered':
            return MessageStatus.DELIVERED
        # Ignore other `Action` values.
        return None

    def get_message_id(self, message: MailMessage) -> Optional[str]:
        """Return a Message-ID of the email object."""
        # If it's a generic reply.
        message_id = cast(Optional[str], message.obj.get('In-Reply-To', message.obj.get('References')))
        if message_id:
            return message_id
        if not message.obj.is_multipart():
            # Proper DSN messages are multipart. This is not the case.
            return None
        # Otherwise expect it to be a DSN. Check payload for the original message.
        for part in message.obj.get_payload():
            if part.get_content_type() == MIMETYPE_MESSAGE:
                # The original message as is.
                if 'Message-ID' in part:
                    return cast(str, part['Message-ID'])
            elif part.get_content_type() == MIMETYPE_MESSAGE_HEADERS:
                # The original headers are in the content of the message. Not an attachment.
                returned = message_from_string(part.get_payload())
                if 'Message-ID' in returned:
                    return cast(str, returned['Message-ID'])
        return None

    def get_dsn_fields(self, message: MailMessage) -> Dict[str, Any]:
        """Return a selected fields from delivery status."""
        if not self.is_dsn(message):
            # Not a valid delivery status
            return {}

        result = {}
        dsn_parts = [part for part in message.obj.get_payload() if part.get_content_type() == MIMETYPE_DSN]
        dsn_fields = [fields for part in dsn_parts for fields in part.get_payload()]
        # XXX: Collect the fields backwards, to keep the first value of each.
        for fields in reversed(dsn_fields):
            result.update({key: value for key, value in fields.items() if key in DSN_FIELD_NAMES})
        return result

    def _get_email(self, session: Session, message_id: str) -> Optional[Union[Email, ArchiveEmail]]:
        """Return an email by Message ID."""
        query = session.query(Email).with_for_update().filter_by(message_id=message_id, status=MessageStatus.SENT)
        try:
            return cast(Email, query.one())
        except NoResultFound:
            # XXX: Doesn't use partitioning. Optimize in case of performance problems.
            query = session.query(ArchiveEmail).with_for_update().filter_by(message_id=message_id,
                                                                            status=ArchiveMessageStatus.SENT)
            try:
                return cast(ArchiveEmail, query.one())
            except NoResultFound:
                return None

    def handle_message(self, message: MailMessage) -> None:
        """Handle an incomming email message.

        Check if the message is reply or a delivery status notification to an existing message and store data.
        """
        status = self.get_status(message)
        if not status:
            # The message is unspecific.
            return
        message_id = self.get_message_id(message)
        if not message_id:
            # No Message-ID found.
            return

        with atomic() as session:
            email = self._get_email(session, message_id)
            if email is None:
                return

            delivery_datetime = datetime.now(UTC)
            # XXX: `message.date` returns dummy date, if Date is empty or not defined :-/
            if message.date_str:
                delivery_datetime = message.date
            delivery_info = self.get_dsn_fields(message)
            if message.subject:
                delivery_info['Subject'] = message.subject
            delivery_message_id = message.obj.get('Message-ID')
            if delivery_message_id is not None:
                delivery_info['Message-ID'] = delivery_message_id
            email.status = status
            email.update_datetime = datetime.now(UTC)
            email.delivery_datetime = delivery_datetime
            email.delivery_info = delivery_info

    def check_delivery(self, limit: int = None) -> int:
        """Check email delivery and return number of processed messages.

        Delivery status messages are deleted from the mailbox.
        """
        num = 0
        with self.get_mailbox() as mailbox:
            for message in mailbox.fetch(AND(seen=False, deleted=False), mark_seen=False, limit=limit):
                num += 1
                try:
                    self.handle_message(message)
                    # Mark all handled messages as seen. Mark DSN messages as deleted.
                    flags = [MailMessageFlags.SEEN]
                    if self.is_dsn(message):
                        flags.append(MailMessageFlags.DELETED)
                    mailbox.flag(message.uid, flags, True)
                except Exception as error:
                    _LOGGER.exception("Processing a delivery message %s caused following error: %s", message, error)
            # Delete all deleted message.
            mailbox.expunge()
            return num
