"""Messenger settings management."""
import logging
import os
from logging.config import dictConfig
from typing import Any, Iterable, Tuple, cast

from easysettings import YAMLSettings

try:  # pragma: no cover
    import sentry_sdk
    from sentry_sdk.integrations.sqlalchemy import SqlalchemyIntegration
except ImportError:
    sentry_sdk = None
    SqlalchemyIntegration = None

_LOGGER = logging.getLogger(__name__)
ENVIRON_NAME = 'MESSENGER_CONFIG'
CONFIG_FILES = ('~/.fred/messenger.conf', '/etc/fred/messenger.conf')
DEFAULT_SETTINGS = {
    'clean_days': 30,
    'db_connection': 'sqlite:///:memory:',
    'db_echo': False,
    'delivery_backend': 'messenger.delivery.email.ImapDelivery',
    'delivery_options': {},
    'processors': {},
    'fileman_netloc': None,
    'fred_db_connection': None,
    'logging': None,
    'grpc_options': None,
    'grpc_port': 50051,
    'monitoring_email': None,
    'msgid_domain': None,
    'senders': {},
    'sentry': {},
    'test_db_connection': 'sqlite:///:memory:',
    # Pseudo-setting which holds a command verbosity level.
    '_verbosity': 1,
}


class Settings:
    """Settings manager."""

    def __init__(self) -> None:
        self._settings = None

    def load(self, config_argv: str = None) -> None:
        """Return setting object from the first config file found.

        Raises:
            FileNotFoundError: If the config file is provided, but the file doesn't exist.
            RuntimeError: If settings are already loaded.
        """
        if self._settings is not None:
            raise RuntimeError('Settings already loaded')
        settings = YAMLSettings(DEFAULT_SETTINGS)
        if ENVIRON_NAME in os.environ:
            settings.load(os.environ[ENVIRON_NAME])
        elif config_argv is not None:
            settings.load(config_argv)
        else:
            for file in CONFIG_FILES:
                try:
                    settings.load(os.path.expanduser(file))
                    # Break the cycle if settings are loaded correctly.
                    break
                except FileNotFoundError:
                    # Ignore if the file doesn't exist.
                    continue
                except OSError as error:
                    # Just log the error and skip to next config file.
                    _LOGGER.warning("Configuration file %s returned error: %s", file, error)
        self._settings = settings
        self._setup_logs()

    def _setup_logs(self) -> None:
        """Set up logging systems."""
        # Apply logging configuration
        if self['logging'] is not None:
            dictConfig(self['logging'])
        # Set up Sentry
        if self['sentry'] and self['sentry'].get('dsn'):
            if sentry_sdk is None:
                _LOGGER.warning('Sentry set up, but sentry-sdk not available.')
            else:
                sentry_sdk.init(dsn=self['sentry']['dsn'], integrations=[SqlalchemyIntegration()],
                                environment=self['sentry'].get('environment'), ca_certs=self['sentry'].get('ca_certs'))

    def __getitem__(self, key: str) -> Any:
        if self._settings is None:
            self.load()
        return cast(YAMLSettings, self._settings)[key]

    def __setitem__(self, key: str, value: Any) -> None:
        if self._settings is None:
            self.load()
        cast(YAMLSettings, self._settings)[key] = value

    def items(self) -> Iterable[Tuple[str, Any]]:
        """Return an iterator of pairs of all settings."""
        if self._settings is None:
            self.load()
        return cast(Iterable[Tuple[str, Any]], cast(YAMLSettings, self._settings).items())


class SettingsProxy:
    """Proxy for settings instance.

    Attributes:
        settings: Actual settings
    """

    def __init__(self, settings: Settings):
        self.settings = settings

    def __getattr__(self, name: str) -> Any:
        return getattr(self.settings, name)

    def __getitem__(self, key: str) -> Any:
        return self.settings[key]

    def __setitem__(self, key: str, value: Any) -> Any:
        self.settings[key] = value


SETTINGS = SettingsProxy(Settings())
