"""Various utility functions."""
import re
import ssl
from email.utils import parseaddr
from importlib import import_module
from os import PathLike
from typing import Any, Optional, Union

import phonenumbers

EMAIL_ADDRESS_RE = re.compile(r'\S+@\S+')


def normalize_email_address(value: str) -> Optional[str]:
    """Return normalized email address, without real name and lowercased, or None if input is not an email address."""
    parsed_email = parseaddr(value)
    bare_email = parsed_email[1].lower()
    if EMAIL_ADDRESS_RE.match(bare_email):
        return bare_email
    else:
        return None


def normalize_phone_number(value: str) -> Optional[str]:
    """Return normalized phone number in E164 format or None if input is not a phone number."""
    try:
        parsed = phonenumbers.parse(value)
    except phonenumbers.NumberParseException:
        return None
    return phonenumbers.format_number(parsed, phonenumbers.PhoneNumberFormat.E164)


def create_ssl_context(certfile: Union[PathLike, str] = None, keyfile: Union[PathLike, str] = None) -> ssl.SSLContext:
    """Create and return a SSL context.

    Arguments:
        certfile: Path to a file with a client certificate.
        keyfile: Path to a file with a private key.
    """
    context = ssl.create_default_context()
    if certfile:
        context.load_cert_chain(certfile, keyfile)
    return context


def import_path(path: str) -> Any:
    """Import a dotted path and return the attribute or class defined.

    Raises:
        ImportError: If the import fails.
    """
    try:
        module_name, class_name = path.rsplit('.', 1)
    except ValueError as error:
        raise ImportError("'{}' is not a valid dotted path.".format(path)) from error

    try:
        module = import_module(module_name)
    except ValueError as error:
        raise ImportError(str(error)) from error

    try:
        return getattr(module, class_name)
    except AttributeError as error:
        raise ImportError("'{}' does not have a class or attribute '{}'.".format(module_name, class_name)) from error


# TODO: Make first argument positional only, once Python 3.8 is minimal version.
def load_instance(_path: str, **kwargs: Any) -> Any:
    """Return instance of class specified by a dotted path and keyword arguments."""
    cls = import_path(_path)
    return cls(**kwargs)
